#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import os
import json
import logging

from typing import Optional

from .. import api, crypto, utils
from ..params import KeeperParams
from .base import suppress_exit, raise_parse_exception, user_choice
from .enterprise_common import EnterpriseCommand
from ..display import bcolors

transfer_user_parser = argparse.ArgumentParser(prog='transfer-user', description='Transfer user account(s).')
transfer_user_parser.add_argument('-f', '--force', dest='force', action='store_true', help='do not prompt for confirmation')
transfer_user_parser.add_argument('--target-user', dest='target_user', action='store', help='email to transfer user(s) to')
transfer_user_parser.add_argument('email', type=str, nargs='+', metavar="user@company.com OR @filename",
                                  help='User account email/ID or File containing account mappings. Use @filename to indicate using mapping file. ' +
                                       'File content: from_account -> to_account')
transfer_user_parser.error = raise_parse_exception
transfer_user_parser.exit = suppress_exit


class EnterpriseTransferUserCommand(EnterpriseCommand):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return transfer_user_parser

    def execute(self, params, **kwargs):
        user_lookup = {}
        if 'users' in params.enterprise:
            for u in params.enterprise['users']:
                user_lookup[str(u['enterprise_user_id'])] = u

                if 'username' in u:
                    user_lookup[u['username'].lower()] = u
                else:
                    logging.debug('WARNING: username is missing from the user id=%s, obj=%s', u['enterprise_user_id'], u)

        transfer_map = {}
        skipped_emails = set()

        def verify_user(username):   # type: (str) -> Optional[str]
            username = username.lower().strip()
            if username in skipped_emails:
                return None
            if username in user_lookup:
                enterprise_user = user_lookup[username]
                if enterprise_user['status'] == 'active':
                    return enterprise_user['username']
                else:
                    logging.warning('\"%s\" is a pending account. Skipping...', username)
                    return None
            else:
                logging.warning('\"%s\" is not a known user account. Skipping...', username)
                return None

        target_user = kwargs['target_user']
        if target_user:
            target_user = verify_user(target_user)

        if 'email' in kwargs:
            for email in kwargs['email']:   # type: str
                if email.startswith('@'):
                    filename = email[1:]
                    if os.path.exists(filename):
                        with open(filename, 'r') as f:
                            lines = f.readlines()
                        for line in lines:
                            if not line:
                                continue
                            line = line.strip()
                            if line[0] in {'#', ';', '-'}:
                                continue
                            p = line.partition('->')
                            if p[1] != '->':
                                p = line.partition('<-')
                                if p[1] != '<-':
                                    p = line.partition('=')
                                    if p[1] != '=':
                                        p = line.partition(' ')
                            if p[2]:
                                user1 = verify_user(p[0])
                                if user1:
                                    user2 = verify_user(p[2])
                                    if user2:
                                        if p[1] == '<-':
                                            from_user = user2
                                            to_user = user1
                                        else:
                                            from_user = user1
                                            to_user = user2
                                        if to_user not in transfer_map:
                                            transfer_map[to_user] = set()
                                        transfer_map[to_user].add(from_user)
                            else:
                                logging.warning('File \"%s\": invalid mapping \"line\". Skipping...', filename)
                    else:
                        logging.warning('File \"%s\" does not exist. Skipping...', filename)
                else:
                    email = verify_user(email)
                    if email and target_user:
                        if target_user not in transfer_map:
                            transfer_map[target_user] = set()
                        transfer_map[target_user].add(email)

        if len(transfer_map) == 0:
            logging.warning('No user accounts to transfer')
            return

        targets = set()
        sources = set()
        for target in transfer_map:
            targets.add(target)
            sources.update(transfer_map[target])
        to_delete = targets.intersection(sources)
        if len(to_delete) > 0:
            for email in to_delete:
                logging.warning('User account \"%s\" appears as both source and target for account transfer', email)
                if email in transfer_map:
                    del transfer_map[email]
            for target in list(transfer_map.keys()):
                transfer_map[target].difference_update(to_delete)
                if len(transfer_map[target]) == 0:
                    del transfer_map[target]

        if len(transfer_map) == 0:
            logging.warning('No user accounts to transfer')
            return

        sources.clear()
        to_delete.clear()
        for target in transfer_map:
            targets = transfer_map[target].intersection(sources)
            if len(targets) > 0:
                to_delete.update(targets)
            sources.update(transfer_map[target])

        if len(to_delete) > 0:
            for email in to_delete:
                logging.warning('User account \"%s\" cannot be moved to multiple targets.', email)
            for target in list(transfer_map.keys()):
                transfer_map[target].difference_update(to_delete)
                if len(transfer_map[target]) == 0:
                    del transfer_map[target]

        if len(transfer_map) == 0:
            logging.warning('No user accounts to transfer')
            return

        targets.clear()
        sources.clear()
        for target in transfer_map:
            targets.add(target)
            sources.update(transfer_map[target])

        if not kwargs.get('force', False):
            answer = user_choice(
                bcolors.FAIL + bcolors.BOLD + '\nALERT!\n' + bcolors.ENDC +
                'This action cannot be undone.\n\n' +
                f'Do you want to proceed with transferring {len(sources)} account(s)?', 'yn', 'n')
            if answer.lower() != 'y':
                return

        lock_rq = []
        for email in sources:
            user = user_lookup[email]
            lock = user.get('lock', 0)
            if lock != 1:
                lock_rq.append({
                    'command': 'enterprise_user_lock',
                    'enterprise_user_id': user['enterprise_user_id'],
                    'lock': 'locked'
                })
        if len(lock_rq) > 0:
            logging.info('Locking active users.')
            api.execute_batch(params, lock_rq)

        public_keys = {}
        for target_user in list(transfer_map.keys()):
            target_public_key = self.get_public_key(params, target_user)
            if target_public_key:
                public_keys[target_user] = target_public_key
            else:
                logging.warning('Failed to get user \"%s\" public key', target_user)
                del transfer_map[target_user]

        for target_user in transfer_map:
            target_public_key = public_keys[target_user]
            for email in transfer_map[target_user]:
                logging.info('Transferring %s account to %s ...', email, target_user)
                self.transfer_user_account(params, email, target_user, target_public_key)

        api.query_enterprise(params)
        params.sync_data = True

    @staticmethod
    def transfer_user_account(params, username, target_user, target_public_key):
        # type: (KeeperParams, str, str, any) -> bool
        rq = {
            'command': 'pre_account_transfer',
            'target_username': username,
        }
        try:
            rs = api.communicate(params, rq)
            tree_key = params.enterprise['unencrypted_tree_key']
            role_key = None
            if 'role_key' in rs:
                role_key = utils.base64_url_decode(rs['role_key'])
                role_key = crypto.decrypt_rsa(role_key, params.rsa_key)
            elif 'role_key_id' in rs:
                role_key_id = rs['role_key_id']
                if 'role_keys2' in params.enterprise:
                    key2 = next((x for x in params.enterprise['role_keys2'] if x['role_id'] == role_key_id), None)
                    if key2:
                        role_key = utils.base64_url_decode(key2['role_key'])
                        role_key = crypto.decrypt_aes_v2(role_key, tree_key)
            if not role_key:
                raise Exception('Cannot resolve Account Transfer role key')
            role_private_key = utils.base64_url_decode(rs['role_private_key']) if 'role_private_key' in rs else None
            role_private_key = crypto.decrypt_aes_v1(role_private_key, role_key)
            role_private_key = crypto.load_rsa_private_key(role_private_key)
            transfer_key = utils.base64_url_decode(rs['transfer_key'])
            transfer_key = crypto.decrypt_rsa(transfer_key, role_private_key)
            user_rsa_private_key = utils.base64_url_decode(rs['user_private_key']) if 'user_private_key' in rs else None
            if user_rsa_private_key:
                user_rsa_private_key = crypto.decrypt_aes_v1(user_rsa_private_key, transfer_key)
                user_rsa_private_key = crypto.load_rsa_private_key(user_rsa_private_key)
            user_ecc_private_key = utils.base64_url_decode(rs['user_ecc_private_key']) if 'user_ecc_private_key' in rs else None
            if user_ecc_private_key:
                user_ecc_private_key = crypto.decrypt_aes_v2(user_ecc_private_key, transfer_key)
                user_ecc_private_key = crypto.load_ec_private_key(user_ecc_private_key)

            rqt = {
                'command': 'transfer_and_delete_user',
                'from_user': username,
                'to_user': target_user
            }
            if 'record_keys' in rs:
                rqt['record_keys'] = []
                rqt['corrupted_record_keys'] = []
                for rk in rs['record_keys']:
                    record_uid = rk['record_uid']
                    try:
                        record_key = utils.base64_url_decode(rk['record_key'])
                        record_key_type = rk.get('record_key_type', 1)
                        if record_key_type == 1:
                            record_key = crypto.decrypt_aes_v1(record_key, transfer_key)
                        elif record_key_type == 2:
                            record_key = crypto.decrypt_rsa(record_key, user_rsa_private_key)
                        elif record_key_type == 3:
                            record_key = crypto.decrypt_aes_v2(record_key, transfer_key)
                        elif record_key_type == 4:
                            record_key = crypto.decrypt_ec(record_key, user_ecc_private_key)
                        else:
                            raise Exception(f'Unsupported record key type')

                        rqt['record_keys'].append({
                            'record_uid': record_uid,
                            'record_key': utils.base64_url_encode(crypto.encrypt_rsa(record_key, target_public_key))
                        })
                    except Exception as e:
                        logging.debug('Corrupted record key: %s: %s', record_uid, e)
                        rqt['corrupted_record_keys'].append(rk)

            if 'shared_folder_keys' in rs:
                rqt['shared_folder_keys'] = []
                rqt['corrupted_shared_folder_keys'] = []
                for sfk in rs['shared_folder_keys']:
                    shared_folder_uid = sfk['shared_folder_uid']
                    try:
                        shared_folder_key = utils.base64_url_decode(sfk['shared_folder_key'])
                        shared_folder_key_type = sfk.get('shared_folder_key_type', 1)
                        if shared_folder_key_type == 1:
                            shared_folder_key = crypto.decrypt_aes_v1(shared_folder_key, transfer_key)
                        elif shared_folder_key_type == 2:
                            shared_folder_key = crypto.decrypt_rsa(shared_folder_key, user_rsa_private_key)
                        elif shared_folder_key_type == 3:
                            shared_folder_key = crypto.decrypt_aes_v2(shared_folder_key, transfer_key)
                        elif shared_folder_key_type == 4:
                            shared_folder_key = crypto.decrypt_ec(shared_folder_key, user_ecc_private_key)
                        else:
                            raise Exception(f'Unsupported shared folder key type')

                        rqt['shared_folder_keys'].append({
                            'shared_folder_uid': shared_folder_uid,
                            'shared_folder_key': utils.base64_url_encode(crypto.encrypt_rsa(shared_folder_key, target_public_key)),
                        })
                    except Exception as e:
                        logging.debug('Corrupted shared folder key: %s: %s', shared_folder_uid, e)
                        rqt['corrupted_shared_folder_keys'].append(sfk)

            if 'team_keys' in rs:
                rqt['team_keys'] = []
                rqt['corrupted_team_keys'] = []
                for tk in rs['team_keys']:
                    team_uid = tk['team_uid']
                    try:
                        team_key = utils.base64_url_decode(tk['team_key'])
                        team_key_type = tk.get('team_key_type', 1)
                        if team_key_type == 1:
                            team_key = crypto.decrypt_aes_v1(team_key, transfer_key)
                        elif team_key_type == 2:
                            team_key = crypto.decrypt_rsa(team_key, user_rsa_private_key)
                        elif team_key_type == 3:
                            team_key = crypto.decrypt_aes_v2(team_key, transfer_key)
                        elif team_key_type == 4:
                            team_key = crypto.decrypt_ec(team_key, user_ecc_private_key)
                        else:
                            raise Exception(f'Unsupported team key type')
                        rqt['team_keys'].append({
                            'team_uid': team_uid,
                            'team_key': utils.base64_url_encode(crypto.encrypt_rsa(team_key, target_public_key)),
                        })
                    except Exception as e:
                        logging.debug('Corrupted team key: %s: %s', team_uid, e)
                        rqt['corrupted_team_keys'].append(tk)

            if 'user_folder_keys' in rs:
                rqt['user_folder_keys'] = []
                rqt['corrupted_user_folder_keys'] = []
                folder_key = utils.generate_aes_key()
                folder_data = json.dumps({
                    'name': f'Transfer from {username}'
                }).encode('utf-8')
                folder_data = crypto.encrypt_aes_v1(folder_data, folder_key)
                rqt['user_folder_transfer'] = {
                    'transfer_folder_uid': utils.generate_uid(),
                    'transfer_folder_key': utils.base64_url_encode(crypto.encrypt_rsa(folder_key, target_public_key)),
                    'transfer_folder_data': utils.base64_url_encode(folder_data)
                }
                for ufk in rs['user_folder_keys']:
                    user_folder_uid = ufk['user_folder_uid']
                    try:
                        user_folder_key = utils.base64_url_decode(ufk['user_folder_key'])
                        user_folder_key_type = ufk.get('user_folder_key_type', 1)
                        if user_folder_key_type == 1:
                            user_folder_key = crypto.decrypt_aes_v1(user_folder_key, transfer_key)
                        elif user_folder_key_type == 2:
                            user_folder_key = crypto.decrypt_rsa(user_folder_key, user_rsa_private_key)
                        elif user_folder_key_type == 3:
                            user_folder_key = crypto.decrypt_aes_v2(user_folder_key, transfer_key)
                        elif user_folder_key_type == 4:
                            user_folder_key = crypto.decrypt_ec(user_folder_key, user_ecc_private_key)
                        else:
                            raise Exception(f'Unsupported user folder key type')
                        rqt['user_folder_keys'].append({
                            'user_folder_uid': user_folder_uid,
                            'user_folder_key': utils.base64_url_encode(crypto.encrypt_rsa(user_folder_key, target_public_key)),
                        })
                    except Exception as e:
                        logging.debug('Corrupted user folder key: %s: %s', user_folder_uid, e)
                        rqt['corrupted_user_folder_keys'].append(ufk)

            api.communicate(params, rqt)
            result = True
            logging.info('%s: account is transferred', username)
            if 'record_keys' in rqt:
                rec_num = len(rqt['record_keys'])
                if rec_num > 0:
                    logging.info(f'{"Records":>16} : {rec_num}')
            if 'shared_folder_keys' in rqt:
                sf_num = len(rqt['shared_folder_keys'])
                if sf_num > 0:
                    logging.info(f'{"SharedFolders":>16} : {sf_num}')

        except Exception as e:
            result = False
            logging.warning('Failed to transfer %s account: %s', username, e)

        return result




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: automator.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import ssocloud_pb2 as ssocloud__pb2
from . import enterprise_pb2 as enterprise__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x61utomator.proto\x12\tAutomator\x1a\x0essocloud.proto\x1a\x10\x65nterprise.proto\"\xbf\x02\n\x15\x41utomatorSettingValue\x12\x11\n\tsettingId\x18\x01 \x01(\x03\x12\x15\n\rsettingTypeId\x18\x02 \x01(\x05\x12\x12\n\nsettingTag\x18\x03 \x01(\t\x12\x13\n\x0bsettingName\x18\x04 \x01(\t\x12\x14\n\x0csettingValue\x18\x05 \x01(\t\x12$\n\x08\x64\x61taType\x18\x06 \x01(\x0e\x32\x12.SsoCloud.DataType\x12\x14\n\x0clastModified\x18\x07 \x01(\t\x12\x10\n\x08\x66romFile\x18\x08 \x01(\x08\x12\x11\n\tencrypted\x18\t \x01(\x08\x12\x0f\n\x07\x65ncoded\x18\n \x01(\x08\x12\x10\n\x08\x65\x64itable\x18\x0b \x01(\x08\x12\x12\n\ntranslated\x18\x0c \x01(\x08\x12\x13\n\x0buserVisible\x18\r \x01(\x08\x12\x10\n\x08required\x18\x0e \x01(\x08\"\xb5\x02\n\x14\x41pproveDeviceRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12O\n\x1dssoAuthenticationProtocolType\x18\x02 \x01(\x0e\x32(.Automator.SsoAuthenticationProtocolType\x12\x13\n\x0b\x61uthMessage\x18\x03 \x01(\t\x12\r\n\x05\x65mail\x18\x04 \x01(\t\x12\x17\n\x0f\x64\x65vicePublicKey\x18\x05 \x01(\x0c\x12\x1c\n\x14serverEccPublicKeyId\x18\x06 \x01(\x05\x12\x1c\n\x14userEncryptedDataKey\x18\x07 \x01(\x0c\x12>\n\x18userEncryptedDataKeyType\x18\x08 \x01(\x0e\x32\x1c.Enterprise.EncryptedKeyType\"\xc7\x01\n\x0cSetupRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x1c\n\x14serverEccPublicKeyId\x18\x02 \x01(\x05\x12\x31\n\x0e\x61utomatorState\x18\x03 \x01(\x0e\x32\x19.Automator.AutomatorState\x12\'\n\x1f\x65ncryptedEnterprisePrivateEcKey\x18\x04 \x01(\x0c\x12(\n encryptedEnterprisePrivateRsaKey\x18\x05 \x01(\x0c\"B\n\rStatusRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x1c\n\x14serverEccPublicKeyId\x18\x02 \x01(\x05\"\xef\x03\n\x11InitializeRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x13\n\x0bidpMetadata\x18\x02 \x01(\t\x12\x1d\n\x15idpSigningCertificate\x18\x03 \x01(\x0c\x12\x13\n\x0bssoEntityId\x18\x04 \x01(\t\x12\x14\n\x0c\x65mailMapping\x18\x05 \x01(\t\x12\x18\n\x10\x66irstnameMapping\x18\x06 \x01(\t\x12\x17\n\x0flastnameMapping\x18\x07 \x01(\t\x12\x10\n\x08\x64isabled\x18\x08 \x01(\x08\x12\x1c\n\x14serverEccPublicKeyId\x18\t \x01(\x05\x12\x0e\n\x06\x63onfig\x18\n \x01(\x0c\x12\x0f\n\x07sslMode\x18\x0b \x01(\t\x12\x14\n\x0cpersistState\x18\x0c \x01(\x08\x12\x17\n\x0f\x64isableSniCheck\x18\r \x01(\x08\x12\x1e\n\x16sslCertificateFilename\x18\x0e \x01(\t\x12\"\n\x1asslCertificateFilePassword\x18\x0f \x01(\t\x12!\n\x19sslCertificateKeyPassword\x18\x10 \x01(\t\x12\x1e\n\x16sslCertificateContents\x18\x11 \x01(\x0c\x12\x15\n\rautomatorHost\x18\x12 \x01(\t\x12\x15\n\rautomatorPort\x18\x13 \x01(\t\"\xa6\x02\n\x16NotInitializedResponse\x12 \n\x18\x61utomatorTransmissionKey\x18\x01 \x01(\x0c\x12\x1a\n\x12signingCertificate\x18\x02 \x01(\x0c\x12\"\n\x1asigningCertificateFilename\x18\x03 \x01(\t\x12\"\n\x1asigningCertificatePassword\x18\x04 \x01(\t\x12\x1a\n\x12signingKeyPassword\x18\x05 \x01(\t\x12>\n\x18signingCertificateFormat\x18\x06 \x01(\x0e\x32\x1c.Automator.CertificateFormat\x12\x1a\n\x12\x61utomatorPublicKey\x18\x07 \x01(\x0c\x12\x0e\n\x06\x63onfig\x18\x08 \x01(\x0c\"\xf9\x02\n\x11\x41utomatorResponse\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x0f\n\x07\x65nabled\x18\x02 \x01(\x08\x12\x11\n\ttimestamp\x18\x03 \x01(\x03\x12\x39\n\rapproveDevice\x18\x04 \x01(\x0b\x32 .Automator.ApproveDeviceResponseH\x00\x12+\n\x06status\x18\x05 \x01(\x0b\x32\x19.Automator.StatusResponseH\x00\x12;\n\x0enotInitialized\x18\x06 \x01(\x0b\x32!.Automator.NotInitializedResponseH\x00\x12)\n\x05\x65rror\x18\x07 \x01(\x0b\x32\x18.Automator.ErrorResponseH\x00\x12\x31\n\x0e\x61utomatorState\x18\x08 \x01(\x0e\x32\x19.Automator.AutomatorState\x12\x1c\n\x14\x61utomatorPublicEcKey\x18\t \x01(\x0c\x42\n\n\x08response\"X\n\x15\x41pproveDeviceResponse\x12\x10\n\x08\x61pproved\x18\x01 \x01(\x08\x12\x1c\n\x14\x65ncryptedUserDataKey\x18\x02 \x01(\x0c\x12\x0f\n\x07message\x18\x03 \x01(\t\"\xc4\x02\n\x0eStatusResponse\x12\x13\n\x0binitialized\x18\x01 \x01(\x08\x12\x18\n\x10\x65nabledTimestamp\x18\x02 \x01(\x03\x12\x1c\n\x14initializedTimestamp\x18\x03 \x01(\x03\x12\x18\n\x10updatedTimestamp\x18\x04 \x01(\x03\x12\x1f\n\x17numberOfDevicesApproved\x18\x05 \x01(\x03\x12\x1d\n\x15numberOfDevicesDenied\x18\x06 \x01(\x03\x12\x16\n\x0enumberOfErrors\x18\x07 \x01(\x03\x12 \n\x18sslCertificateExpiration\x18\x08 \x01(\x03\x12\x41\n\x16notInitializedResponse\x18\t \x01(\x0b\x32!.Automator.NotInitializedResponse\x12\x0e\n\x06\x63onfig\x18\n \x01(\x0c\" \n\rErrorResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"X\n\x08LogEntry\x12\x12\n\nserverTime\x18\x01 \x01(\t\x12\x14\n\x0cmessageLevel\x18\x02 \x01(\t\x12\x11\n\tcomponent\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\t\"b\n\rAdminResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12/\n\rautomatorInfo\x18\x03 \x03(\x0b\x32\x18.Automator.AutomatorInfo\"\xdd\x02\n\rAutomatorInfo\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x0e\n\x06nodeId\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x04 \x01(\x08\x12\x0b\n\x03url\x18\x05 \x01(\t\x12\x32\n\x0f\x61utomatorSkills\x18\x06 \x03(\x0b\x32\x19.Automator.AutomatorSkill\x12@\n\x16\x61utomatorSettingValues\x18\x07 \x03(\x0b\x32 .Automator.AutomatorSettingValue\x12)\n\x06status\x18\x08 \x01(\x0b\x32\x19.Automator.StatusResponse\x12\'\n\nlogEntries\x18\t \x03(\x0b\x32\x13.Automator.LogEntry\x12\x31\n\x0e\x61utomatorState\x18\n \x01(\x0e\x32\x19.Automator.AutomatorState\"e\n\x1b\x41\x64minCreateAutomatorRequest\x12\x0e\n\x06nodeId\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12(\n\x05skill\x18\x03 \x01(\x0b\x32\x19.Automator.AutomatorSkill\"2\n\x1b\x41\x64minDeleteAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"1\n\x1f\x41\x64minGetAutomatorsOnNodeRequest\x12\x0e\n\x06nodeId\x18\x01 \x01(\x03\">\n&AdminGetAutomatorsForEnterpriseRequest\x12\x14\n\x0c\x65nterpriseId\x18\x01 \x01(\x05\"/\n\x18\x41\x64minGetAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"C\n\x1b\x41\x64minEnableAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x0f\n\x07\x65nabled\x18\x02 \x01(\x08\"\xc8\x01\n\x19\x41\x64minEditAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x03 \x01(\x08\x12\x0b\n\x03url\x18\x04 \x01(\t\x12(\n\nskillTypes\x18\x05 \x03(\x0e\x32\x14.Automator.SkillType\x12@\n\x16\x61utomatorSettingValues\x18\x06 \x03(\x0b\x32 .Automator.AutomatorSettingValue\"\xb7\x01\n\x1a\x41\x64minSetupAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x31\n\x0e\x61utomatorState\x18\x02 \x01(\x0e\x32\x19.Automator.AutomatorState\x12\'\n\x1f\x65ncryptedEcEnterprisePrivateKey\x18\x03 \x01(\x0c\x12(\n encryptedRsaEnterprisePrivateKey\x18\x04 \x01(\x0c\"\xa5\x01\n\x1b\x41\x64minSetupAutomatorResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x13\n\x0b\x61utomatorId\x18\x03 \x01(\x03\x12\x31\n\x0e\x61utomatorState\x18\x04 \x01(\x0e\x32\x19.Automator.AutomatorState\x12\x1c\n\x14\x61utomatorEcPublicKey\x18\x05 \x01(\x0c\"2\n\x1b\x41\x64minAutomatorSkillsRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"_\n\x0e\x41utomatorSkill\x12\'\n\tskillType\x18\x01 \x01(\x0e\x32\x14.Automator.SkillType\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x16\n\x0etranslatedName\x18\x03 \x01(\t\"t\n\x1c\x41\x64minAutomatorSkillsResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x32\n\x0f\x61utomatorSkills\x18\x03 \x03(\x0b\x32\x19.Automator.AutomatorSkill\"1\n\x1a\x41\x64minResetAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"6\n\x1f\x41\x64minInitializeAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"/\n\x18\x41\x64minAutomatorLogRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"4\n\x1d\x41\x64minAutomatorLogClearRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03*@\n\x1dSsoAuthenticationProtocolType\x12\x14\n\x10UNKNOWN_PROTOCOL\x10\x00\x12\t\n\x05SAML2\x10\x01*<\n\x11\x43\x65rtificateFormat\x12\x12\n\x0eUNKNOWN_FORMAT\x10\x00\x12\n\n\x06PKCS12\x10\x01\x12\x07\n\x03JKS\x10\x02*K\n\tSkillType\x12\x16\n\x12UNKNOWN_SKILL_TYPE\x10\x00\x12\x13\n\x0f\x44\x45VICE_APPROVAL\x10\x01\x12\x11\n\rTEAM_APPROVAL\x10\x02*\x87\x01\n\x0e\x41utomatorState\x12\x11\n\rUNKNOWN_STATE\x10\x00\x12\x0b\n\x07RUNNING\x10\x01\x12\t\n\x05\x45RROR\x10\x02\x12\x18\n\x14NEEDS_INITIALIZATION\x10\x03\x12\x17\n\x13NEEDS_CRYPTO_STEP_1\x10\x04\x12\x17\n\x13NEEDS_CRYPTO_STEP_2\x10\x05\x42%\n\x18\x63om.keepersecurity.protoB\tAutomatorb\x06proto3')

_SSOAUTHENTICATIONPROTOCOLTYPE = DESCRIPTOR.enum_types_by_name['SsoAuthenticationProtocolType']
SsoAuthenticationProtocolType = enum_type_wrapper.EnumTypeWrapper(_SSOAUTHENTICATIONPROTOCOLTYPE)
_CERTIFICATEFORMAT = DESCRIPTOR.enum_types_by_name['CertificateFormat']
CertificateFormat = enum_type_wrapper.EnumTypeWrapper(_CERTIFICATEFORMAT)
_SKILLTYPE = DESCRIPTOR.enum_types_by_name['SkillType']
SkillType = enum_type_wrapper.EnumTypeWrapper(_SKILLTYPE)
_AUTOMATORSTATE = DESCRIPTOR.enum_types_by_name['AutomatorState']
AutomatorState = enum_type_wrapper.EnumTypeWrapper(_AUTOMATORSTATE)
UNKNOWN_PROTOCOL = 0
SAML2 = 1
UNKNOWN_FORMAT = 0
PKCS12 = 1
JKS = 2
UNKNOWN_SKILL_TYPE = 0
DEVICE_APPROVAL = 1
TEAM_APPROVAL = 2
UNKNOWN_STATE = 0
RUNNING = 1
ERROR = 2
NEEDS_INITIALIZATION = 3
NEEDS_CRYPTO_STEP_1 = 4
NEEDS_CRYPTO_STEP_2 = 5


_AUTOMATORSETTINGVALUE = DESCRIPTOR.message_types_by_name['AutomatorSettingValue']
_APPROVEDEVICEREQUEST = DESCRIPTOR.message_types_by_name['ApproveDeviceRequest']
_SETUPREQUEST = DESCRIPTOR.message_types_by_name['SetupRequest']
_STATUSREQUEST = DESCRIPTOR.message_types_by_name['StatusRequest']
_INITIALIZEREQUEST = DESCRIPTOR.message_types_by_name['InitializeRequest']
_NOTINITIALIZEDRESPONSE = DESCRIPTOR.message_types_by_name['NotInitializedResponse']
_AUTOMATORRESPONSE = DESCRIPTOR.message_types_by_name['AutomatorResponse']
_APPROVEDEVICERESPONSE = DESCRIPTOR.message_types_by_name['ApproveDeviceResponse']
_STATUSRESPONSE = DESCRIPTOR.message_types_by_name['StatusResponse']
_ERRORRESPONSE = DESCRIPTOR.message_types_by_name['ErrorResponse']
_LOGENTRY = DESCRIPTOR.message_types_by_name['LogEntry']
_ADMINRESPONSE = DESCRIPTOR.message_types_by_name['AdminResponse']
_AUTOMATORINFO = DESCRIPTOR.message_types_by_name['AutomatorInfo']
_ADMINCREATEAUTOMATORREQUEST = DESCRIPTOR.message_types_by_name['AdminCreateAutomatorRequest']
_ADMINDELETEAUTOMATORREQUEST = DESCRIPTOR.message_types_by_name['AdminDeleteAutomatorRequest']
_ADMINGETAUTOMATORSONNODEREQUEST = DESCRIPTOR.message_types_by_name['AdminGetAutomatorsOnNodeRequest']
_ADMINGETAUTOMATORSFORENTERPRISEREQUEST = DESCRIPTOR.message_types_by_name['AdminGetAutomatorsForEnterpriseRequest']
_ADMINGETAUTOMATORREQUEST = DESCRIPTOR.message_types_by_name['AdminGetAutomatorRequest']
_ADMINENABLEAUTOMATORREQUEST = DESCRIPTOR.message_types_by_name['AdminEnableAutomatorRequest']
_ADMINEDITAUTOMATORREQUEST = DESCRIPTOR.message_types_by_name['AdminEditAutomatorRequest']
_ADMINSETUPAUTOMATORREQUEST = DESCRIPTOR.message_types_by_name['AdminSetupAutomatorRequest']
_ADMINSETUPAUTOMATORRESPONSE = DESCRIPTOR.message_types_by_name['AdminSetupAutomatorResponse']
_ADMINAUTOMATORSKILLSREQUEST = DESCRIPTOR.message_types_by_name['AdminAutomatorSkillsRequest']
_AUTOMATORSKILL = DESCRIPTOR.message_types_by_name['AutomatorSkill']
_ADMINAUTOMATORSKILLSRESPONSE = DESCRIPTOR.message_types_by_name['AdminAutomatorSkillsResponse']
_ADMINRESETAUTOMATORREQUEST = DESCRIPTOR.message_types_by_name['AdminResetAutomatorRequest']
_ADMININITIALIZEAUTOMATORREQUEST = DESCRIPTOR.message_types_by_name['AdminInitializeAutomatorRequest']
_ADMINAUTOMATORLOGREQUEST = DESCRIPTOR.message_types_by_name['AdminAutomatorLogRequest']
_ADMINAUTOMATORLOGCLEARREQUEST = DESCRIPTOR.message_types_by_name['AdminAutomatorLogClearRequest']
AutomatorSettingValue = _reflection.GeneratedProtocolMessageType('AutomatorSettingValue', (_message.Message,), {
  'DESCRIPTOR' : _AUTOMATORSETTINGVALUE,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AutomatorSettingValue)
  })
_sym_db.RegisterMessage(AutomatorSettingValue)

ApproveDeviceRequest = _reflection.GeneratedProtocolMessageType('ApproveDeviceRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPROVEDEVICEREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.ApproveDeviceRequest)
  })
_sym_db.RegisterMessage(ApproveDeviceRequest)

SetupRequest = _reflection.GeneratedProtocolMessageType('SetupRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETUPREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.SetupRequest)
  })
_sym_db.RegisterMessage(SetupRequest)

StatusRequest = _reflection.GeneratedProtocolMessageType('StatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _STATUSREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.StatusRequest)
  })
_sym_db.RegisterMessage(StatusRequest)

InitializeRequest = _reflection.GeneratedProtocolMessageType('InitializeRequest', (_message.Message,), {
  'DESCRIPTOR' : _INITIALIZEREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.InitializeRequest)
  })
_sym_db.RegisterMessage(InitializeRequest)

NotInitializedResponse = _reflection.GeneratedProtocolMessageType('NotInitializedResponse', (_message.Message,), {
  'DESCRIPTOR' : _NOTINITIALIZEDRESPONSE,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.NotInitializedResponse)
  })
_sym_db.RegisterMessage(NotInitializedResponse)

AutomatorResponse = _reflection.GeneratedProtocolMessageType('AutomatorResponse', (_message.Message,), {
  'DESCRIPTOR' : _AUTOMATORRESPONSE,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AutomatorResponse)
  })
_sym_db.RegisterMessage(AutomatorResponse)

ApproveDeviceResponse = _reflection.GeneratedProtocolMessageType('ApproveDeviceResponse', (_message.Message,), {
  'DESCRIPTOR' : _APPROVEDEVICERESPONSE,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.ApproveDeviceResponse)
  })
_sym_db.RegisterMessage(ApproveDeviceResponse)

StatusResponse = _reflection.GeneratedProtocolMessageType('StatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _STATUSRESPONSE,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.StatusResponse)
  })
_sym_db.RegisterMessage(StatusResponse)

ErrorResponse = _reflection.GeneratedProtocolMessageType('ErrorResponse', (_message.Message,), {
  'DESCRIPTOR' : _ERRORRESPONSE,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.ErrorResponse)
  })
_sym_db.RegisterMessage(ErrorResponse)

LogEntry = _reflection.GeneratedProtocolMessageType('LogEntry', (_message.Message,), {
  'DESCRIPTOR' : _LOGENTRY,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.LogEntry)
  })
_sym_db.RegisterMessage(LogEntry)

AdminResponse = _reflection.GeneratedProtocolMessageType('AdminResponse', (_message.Message,), {
  'DESCRIPTOR' : _ADMINRESPONSE,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminResponse)
  })
_sym_db.RegisterMessage(AdminResponse)

AutomatorInfo = _reflection.GeneratedProtocolMessageType('AutomatorInfo', (_message.Message,), {
  'DESCRIPTOR' : _AUTOMATORINFO,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AutomatorInfo)
  })
_sym_db.RegisterMessage(AutomatorInfo)

AdminCreateAutomatorRequest = _reflection.GeneratedProtocolMessageType('AdminCreateAutomatorRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINCREATEAUTOMATORREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminCreateAutomatorRequest)
  })
_sym_db.RegisterMessage(AdminCreateAutomatorRequest)

AdminDeleteAutomatorRequest = _reflection.GeneratedProtocolMessageType('AdminDeleteAutomatorRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINDELETEAUTOMATORREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminDeleteAutomatorRequest)
  })
_sym_db.RegisterMessage(AdminDeleteAutomatorRequest)

AdminGetAutomatorsOnNodeRequest = _reflection.GeneratedProtocolMessageType('AdminGetAutomatorsOnNodeRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINGETAUTOMATORSONNODEREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminGetAutomatorsOnNodeRequest)
  })
_sym_db.RegisterMessage(AdminGetAutomatorsOnNodeRequest)

AdminGetAutomatorsForEnterpriseRequest = _reflection.GeneratedProtocolMessageType('AdminGetAutomatorsForEnterpriseRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINGETAUTOMATORSFORENTERPRISEREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminGetAutomatorsForEnterpriseRequest)
  })
_sym_db.RegisterMessage(AdminGetAutomatorsForEnterpriseRequest)

AdminGetAutomatorRequest = _reflection.GeneratedProtocolMessageType('AdminGetAutomatorRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINGETAUTOMATORREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminGetAutomatorRequest)
  })
_sym_db.RegisterMessage(AdminGetAutomatorRequest)

AdminEnableAutomatorRequest = _reflection.GeneratedProtocolMessageType('AdminEnableAutomatorRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINENABLEAUTOMATORREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminEnableAutomatorRequest)
  })
_sym_db.RegisterMessage(AdminEnableAutomatorRequest)

AdminEditAutomatorRequest = _reflection.GeneratedProtocolMessageType('AdminEditAutomatorRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINEDITAUTOMATORREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminEditAutomatorRequest)
  })
_sym_db.RegisterMessage(AdminEditAutomatorRequest)

AdminSetupAutomatorRequest = _reflection.GeneratedProtocolMessageType('AdminSetupAutomatorRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINSETUPAUTOMATORREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminSetupAutomatorRequest)
  })
_sym_db.RegisterMessage(AdminSetupAutomatorRequest)

AdminSetupAutomatorResponse = _reflection.GeneratedProtocolMessageType('AdminSetupAutomatorResponse', (_message.Message,), {
  'DESCRIPTOR' : _ADMINSETUPAUTOMATORRESPONSE,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminSetupAutomatorResponse)
  })
_sym_db.RegisterMessage(AdminSetupAutomatorResponse)

AdminAutomatorSkillsRequest = _reflection.GeneratedProtocolMessageType('AdminAutomatorSkillsRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINAUTOMATORSKILLSREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminAutomatorSkillsRequest)
  })
_sym_db.RegisterMessage(AdminAutomatorSkillsRequest)

AutomatorSkill = _reflection.GeneratedProtocolMessageType('AutomatorSkill', (_message.Message,), {
  'DESCRIPTOR' : _AUTOMATORSKILL,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AutomatorSkill)
  })
_sym_db.RegisterMessage(AutomatorSkill)

AdminAutomatorSkillsResponse = _reflection.GeneratedProtocolMessageType('AdminAutomatorSkillsResponse', (_message.Message,), {
  'DESCRIPTOR' : _ADMINAUTOMATORSKILLSRESPONSE,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminAutomatorSkillsResponse)
  })
_sym_db.RegisterMessage(AdminAutomatorSkillsResponse)

AdminResetAutomatorRequest = _reflection.GeneratedProtocolMessageType('AdminResetAutomatorRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINRESETAUTOMATORREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminResetAutomatorRequest)
  })
_sym_db.RegisterMessage(AdminResetAutomatorRequest)

AdminInitializeAutomatorRequest = _reflection.GeneratedProtocolMessageType('AdminInitializeAutomatorRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMININITIALIZEAUTOMATORREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminInitializeAutomatorRequest)
  })
_sym_db.RegisterMessage(AdminInitializeAutomatorRequest)

AdminAutomatorLogRequest = _reflection.GeneratedProtocolMessageType('AdminAutomatorLogRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINAUTOMATORLOGREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminAutomatorLogRequest)
  })
_sym_db.RegisterMessage(AdminAutomatorLogRequest)

AdminAutomatorLogClearRequest = _reflection.GeneratedProtocolMessageType('AdminAutomatorLogClearRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADMINAUTOMATORLOGCLEARREQUEST,
  '__module__' : 'automator_pb2'
  # @@protoc_insertion_point(class_scope:Automator.AdminAutomatorLogClearRequest)
  })
_sym_db.RegisterMessage(AdminAutomatorLogClearRequest)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.keepersecurity.protoB\tAutomator'
  _SSOAUTHENTICATIONPROTOCOLTYPE._serialized_start=4558
  _SSOAUTHENTICATIONPROTOCOLTYPE._serialized_end=4622
  _CERTIFICATEFORMAT._serialized_start=4624
  _CERTIFICATEFORMAT._serialized_end=4684
  _SKILLTYPE._serialized_start=4686
  _SKILLTYPE._serialized_end=4761
  _AUTOMATORSTATE._serialized_start=4764
  _AUTOMATORSTATE._serialized_end=4899
  _AUTOMATORSETTINGVALUE._serialized_start=65
  _AUTOMATORSETTINGVALUE._serialized_end=384
  _APPROVEDEVICEREQUEST._serialized_start=387
  _APPROVEDEVICEREQUEST._serialized_end=696
  _SETUPREQUEST._serialized_start=699
  _SETUPREQUEST._serialized_end=898
  _STATUSREQUEST._serialized_start=900
  _STATUSREQUEST._serialized_end=966
  _INITIALIZEREQUEST._serialized_start=969
  _INITIALIZEREQUEST._serialized_end=1464
  _NOTINITIALIZEDRESPONSE._serialized_start=1467
  _NOTINITIALIZEDRESPONSE._serialized_end=1761
  _AUTOMATORRESPONSE._serialized_start=1764
  _AUTOMATORRESPONSE._serialized_end=2141
  _APPROVEDEVICERESPONSE._serialized_start=2143
  _APPROVEDEVICERESPONSE._serialized_end=2231
  _STATUSRESPONSE._serialized_start=2234
  _STATUSRESPONSE._serialized_end=2558
  _ERRORRESPONSE._serialized_start=2560
  _ERRORRESPONSE._serialized_end=2592
  _LOGENTRY._serialized_start=2594
  _LOGENTRY._serialized_end=2682
  _ADMINRESPONSE._serialized_start=2684
  _ADMINRESPONSE._serialized_end=2782
  _AUTOMATORINFO._serialized_start=2785
  _AUTOMATORINFO._serialized_end=3134
  _ADMINCREATEAUTOMATORREQUEST._serialized_start=3136
  _ADMINCREATEAUTOMATORREQUEST._serialized_end=3237
  _ADMINDELETEAUTOMATORREQUEST._serialized_start=3239
  _ADMINDELETEAUTOMATORREQUEST._serialized_end=3289
  _ADMINGETAUTOMATORSONNODEREQUEST._serialized_start=3291
  _ADMINGETAUTOMATORSONNODEREQUEST._serialized_end=3340
  _ADMINGETAUTOMATORSFORENTERPRISEREQUEST._serialized_start=3342
  _ADMINGETAUTOMATORSFORENTERPRISEREQUEST._serialized_end=3404
  _ADMINGETAUTOMATORREQUEST._serialized_start=3406
  _ADMINGETAUTOMATORREQUEST._serialized_end=3453
  _ADMINENABLEAUTOMATORREQUEST._serialized_start=3455
  _ADMINENABLEAUTOMATORREQUEST._serialized_end=3522
  _ADMINEDITAUTOMATORREQUEST._serialized_start=3525
  _ADMINEDITAUTOMATORREQUEST._serialized_end=3725
  _ADMINSETUPAUTOMATORREQUEST._serialized_start=3728
  _ADMINSETUPAUTOMATORREQUEST._serialized_end=3911
  _ADMINSETUPAUTOMATORRESPONSE._serialized_start=3914
  _ADMINSETUPAUTOMATORRESPONSE._serialized_end=4079
  _ADMINAUTOMATORSKILLSREQUEST._serialized_start=4081
  _ADMINAUTOMATORSKILLSREQUEST._serialized_end=4131
  _AUTOMATORSKILL._serialized_start=4133
  _AUTOMATORSKILL._serialized_end=4228
  _ADMINAUTOMATORSKILLSRESPONSE._serialized_start=4230
  _ADMINAUTOMATORSKILLSRESPONSE._serialized_end=4346
  _ADMINRESETAUTOMATORREQUEST._serialized_start=4348
  _ADMINRESETAUTOMATORREQUEST._serialized_end=4397
  _ADMININITIALIZEAUTOMATORREQUEST._serialized_start=4399
  _ADMININITIALIZEAUTOMATORREQUEST._serialized_end=4453
  _ADMINAUTOMATORLOGREQUEST._serialized_start=4455
  _ADMINAUTOMATORLOGREQUEST._serialized_end=4502
  _ADMINAUTOMATORLOGCLEARREQUEST._serialized_start=4504
  _ADMINAUTOMATORLOGCLEARREQUEST._serialized_end=4556
# @@protoc_insertion_point(module_scope)



from unittest import TestCase, mock
from collections import namedtuple

from data_vault import VaultEnvironment, get_synced_params, get_connected_params
from helper import KeeperApiHelper
from keepercommander import api, generator

vault_env = VaultEnvironment()

PasswordStrength = namedtuple('PasswordStrength', 'length caps lower digits symbols')


class TestPasswordGenerator(TestCase):
    @staticmethod
    def get_password_strength(password):  # type: (str) -> PasswordStrength
        length = len(password)
        caps = 0
        lower = 0
        digits = 0
        symbols = 0

        for ch in password:
            if ch.isalpha():
                if ch.isupper():
                    caps += 1
                else:
                    lower += 1
            elif ch.isdigit():
                digits += 1
            else:
                symbols += 1
        return PasswordStrength(length=length, caps=caps, lower=lower, digits=digits, symbols=symbols)

    def test_generator_exclude(self):
        gen = generator.KeeperPasswordGenerator(length=20, caps=-2, lower=-2, digits=-2, symbols=0)
        self.assertEqual(gen.category_map[4][0], 14)
        password = gen.generate()
        strength = TestPasswordGenerator.get_password_strength(password)
        self.assertEqual(strength.length, 20)
        self.assertEqual(strength.symbols, 0)

    def test_generator_fail(self):
        with self.assertRaises(Exception):
            generator.KeeperPasswordGenerator(length=20, caps=0, lower=0, digits=0, symbols=0)


class TestSearch(TestCase):
    def setUp(self):
        self.communicate_mock = mock.patch('keepercommander.api.communicate').start()
        self.communicate_mock.side_effect = KeeperApiHelper.communicate_command

    def tearDown(self):
        mock.patch.stopall()

    def test_search_records(self):
        params = get_synced_params()

        records = api.search_records(params, '')
        self.assertEqual(len(records), len(params.record_cache))

        records = api.search_records(params, 'RECORD')
        self.assertEqual(len(records), len(params.record_cache))

        records = api.search_records(params, 'Record 1')
        self.assertEqual(len(records), 1)

        records = api.search_records(params, 'INVALID')
        self.assertEqual(len(records), 0)

    def test_search_shared_folders(self):
        params = get_synced_params()

        sfs = api.search_shared_folders(params, '')
        self.assertEqual(len(sfs), len(params.shared_folder_cache))

        sfs = api.search_shared_folders(params, 'folder')
        self.assertEqual(len(sfs), len(params.shared_folder_cache))

        sfs = api.search_shared_folders(params, '1')
        self.assertEqual(len(sfs), 1)

        sfs = api.search_shared_folders(params, 'INVALID')
        self.assertEqual(len(sfs), 0)

    def test_search_teams(self):
        params = get_synced_params()

        teams = api.search_teams(params, '')
        self.assertEqual(len(teams), len(params.team_cache))

        teams = api.search_shared_folders(params, 'team')
        self.assertEqual(len(teams), len(params.shared_folder_cache))

        teams = api.search_shared_folders(params, '1')
        self.assertEqual(len(teams), 1)

        teams = api.search_shared_folders(params, 'INVALID')
        self.assertEqual(len(teams), 0)

    def test_accept_account_transfer_consent(self):
        params = get_connected_params()
        params.settings = {
            'must_perform_account_share_by': '1632370067000',
            'share_account_to': [{
                'role_id': 123456789,
                'public_key': vault_env.encoded_public_key
            }]
        }
        with mock.patch('builtins.print'), mock.patch('builtins.input', return_value='accept'):

            KeeperApiHelper.communicate_expect(['share_account'])
            self.assertTrue(api.accept_account_transfer_consent(params))
            self.assertTrue(KeeperApiHelper.is_expect_empty())

    def test_decline_account_transfer_consent(self):
        params = get_connected_params()
        params.settings = {
            'must_perform_account_share_by': '1632370067000',
            'share_account_to': [{
                'role_id': 123456789,
                'public_key': vault_env.encoded_public_key
            }]
        }
        with mock.patch('builtins.print'), mock.patch('builtins.input', return_value='decline'):
            self.assertFalse(api.accept_account_transfer_consent(params))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: router.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import pam_pb2 as pam__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0crouter.proto\x12\x06Router\x1a\tpam.proto\"r\n\x0eRouterResponse\x12\x30\n\x0cresponseCode\x18\x01 \x01(\x0e\x32\x1a.Router.RouterResponseCode\x12\x14\n\x0c\x65rrorMessage\x18\x02 \x01(\t\x12\x18\n\x10\x65ncryptedPayload\x18\x03 \x01(\x0c\"\xa1\x01\n\x17RouterControllerMessage\x12\x32\n\x0bmessageType\x18\x01 \x01(\x0e\x32\x1d.Router.ControllerMessageType\x12\x12\n\nmessageUid\x18\x02 \x01(\x0c\x12\x15\n\rcontrollerUid\x18\x03 \x01(\x0c\x12\x16\n\x0estreamResponse\x18\x04 \x01(\x08\x12\x0f\n\x07payload\x18\x05 \x01(\x0c\"\x92\x01\n\x0eRouterUserAuth\x12\x17\n\x0ftransmissionKey\x18\x01 \x01(\x0c\x12\x14\n\x0csessionToken\x18\x02 \x01(\x0c\x12\x0e\n\x06userId\x18\x03 \x01(\x05\x12\x18\n\x10\x65nterpriseUserId\x18\x04 \x01(\x03\x12\x12\n\ndeviceName\x18\x05 \x01(\t\x12\x13\n\x0b\x64\x65viceToken\x18\x06 \x01(\x0c\"\xb9\x01\n\x10RouterDeviceAuth\x12\x10\n\x08\x63lientId\x18\x01 \x01(\t\x12\x15\n\rclientVersion\x18\x02 \x01(\t\x12\x14\n\x0c\x65nterpriseId\x18\x03 \x01(\x05\x12\x0e\n\x06nodeId\x18\x04 \x01(\x03\x12\x12\n\ndeviceName\x18\x05 \x01(\t\x12\x13\n\x0b\x64\x65viceToken\x18\x06 \x01(\x0c\x12\x16\n\x0e\x63ontrollerName\x18\x07 \x01(\t\x12\x15\n\rcontrollerUid\x18\x08 \x01(\x0c\"n\n\x14RouterRecordRotation\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x18\n\x10\x63onfigurationUid\x18\x02 \x01(\x0c\x12\x15\n\rcontrollerUid\x18\x03 \x01(\x0c\x12\x12\n\nnoSchedule\x18\x04 \x01(\x08\"E\n\x1cRouterRecordRotationsRequest\x12\x14\n\x0c\x65nterpriseId\x18\x01 \x01(\x05\x12\x0f\n\x07records\x18\x02 \x03(\x0c\"a\n\x1dRouterRecordRotationsResponse\x12/\n\trotations\x18\x01 \x03(\x0b\x32\x1c.Router.RouterRecordRotation\x12\x0f\n\x07hasMore\x18\x02 \x01(\x08\"\xe8\x01\n\x12RouterRotationInfo\x12,\n\x06status\x18\x01 \x01(\x0e\x32\x1c.Router.RouterRotationStatus\x12\x18\n\x10\x63onfigurationUid\x18\x02 \x01(\x0c\x12\x0e\n\x06nodeId\x18\x03 \x01(\x03\x12\x15\n\rcontrollerUid\x18\x04 \x01(\x0c\x12\x16\n\x0e\x63ontrollerName\x18\x05 \x01(\t\x12\x0e\n\x06\x63ookie\x18\x06 \x01(\t\x12\x12\n\nscriptName\x18\x07 \x01(\t\x12\x15\n\rpwdComplexity\x18\x08 \x01(\t\x12\x10\n\x08\x64isabled\x18\t \x01(\x08\"\xb1\x01\n\x1bRouterRecordRotationRequest\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x18\n\x10\x63onfigurationUid\x18\x03 \x01(\x0c\x12\x10\n\x08schedule\x18\x04 \x01(\t\x12\x18\n\x10\x65nterpriseUserId\x18\x05 \x01(\x03\x12\x15\n\rpwdComplexity\x18\x06 \x01(\x0c\x12\x10\n\x08\x64isabled\x18\x07 \x01(\x08*\xb6\x01\n\x12RouterResponseCode\x12\n\n\x06RRC_OK\x10\x00\x12\x15\n\x11RRC_GENERAL_ERROR\x10\x01\x12\x13\n\x0fRRC_NOT_ALLOWED\x10\x02\x12\x13\n\x0fRRC_BAD_REQUEST\x10\x03\x12\x0f\n\x0bRRC_TIMEOUT\x10\x04\x12\x11\n\rRRC_BAD_STATE\x10\x05\x12\x17\n\x13RRC_CONTROLLER_DOWN\x10\x06\x12\x16\n\x12RRC_WRONG_INSTANCE\x10\x07*k\n\x14RouterRotationStatus\x12\x0e\n\nRRS_ONLINE\x10\x00\x12\x13\n\x0fRRS_NO_ROTATION\x10\x01\x12\x15\n\x11RRS_NO_CONTROLLER\x10\x02\x12\x17\n\x13RRS_CONTROLLER_DOWN\x10\x03\x42\"\n\x18\x63om.keepersecurity.protoB\x06Routerb\x06proto3')

_ROUTERRESPONSECODE = DESCRIPTOR.enum_types_by_name['RouterResponseCode']
RouterResponseCode = enum_type_wrapper.EnumTypeWrapper(_ROUTERRESPONSECODE)
_ROUTERROTATIONSTATUS = DESCRIPTOR.enum_types_by_name['RouterRotationStatus']
RouterRotationStatus = enum_type_wrapper.EnumTypeWrapper(_ROUTERROTATIONSTATUS)
RRC_OK = 0
RRC_GENERAL_ERROR = 1
RRC_NOT_ALLOWED = 2
RRC_BAD_REQUEST = 3
RRC_TIMEOUT = 4
RRC_BAD_STATE = 5
RRC_CONTROLLER_DOWN = 6
RRC_WRONG_INSTANCE = 7
RRS_ONLINE = 0
RRS_NO_ROTATION = 1
RRS_NO_CONTROLLER = 2
RRS_CONTROLLER_DOWN = 3


_ROUTERRESPONSE = DESCRIPTOR.message_types_by_name['RouterResponse']
_ROUTERCONTROLLERMESSAGE = DESCRIPTOR.message_types_by_name['RouterControllerMessage']
_ROUTERUSERAUTH = DESCRIPTOR.message_types_by_name['RouterUserAuth']
_ROUTERDEVICEAUTH = DESCRIPTOR.message_types_by_name['RouterDeviceAuth']
_ROUTERRECORDROTATION = DESCRIPTOR.message_types_by_name['RouterRecordRotation']
_ROUTERRECORDROTATIONSREQUEST = DESCRIPTOR.message_types_by_name['RouterRecordRotationsRequest']
_ROUTERRECORDROTATIONSRESPONSE = DESCRIPTOR.message_types_by_name['RouterRecordRotationsResponse']
_ROUTERROTATIONINFO = DESCRIPTOR.message_types_by_name['RouterRotationInfo']
_ROUTERRECORDROTATIONREQUEST = DESCRIPTOR.message_types_by_name['RouterRecordRotationRequest']
RouterResponse = _reflection.GeneratedProtocolMessageType('RouterResponse', (_message.Message,), {
  'DESCRIPTOR' : _ROUTERRESPONSE,
  '__module__' : 'router_pb2'
  # @@protoc_insertion_point(class_scope:Router.RouterResponse)
  })
_sym_db.RegisterMessage(RouterResponse)

RouterControllerMessage = _reflection.GeneratedProtocolMessageType('RouterControllerMessage', (_message.Message,), {
  'DESCRIPTOR' : _ROUTERCONTROLLERMESSAGE,
  '__module__' : 'router_pb2'
  # @@protoc_insertion_point(class_scope:Router.RouterControllerMessage)
  })
_sym_db.RegisterMessage(RouterControllerMessage)

RouterUserAuth = _reflection.GeneratedProtocolMessageType('RouterUserAuth', (_message.Message,), {
  'DESCRIPTOR' : _ROUTERUSERAUTH,
  '__module__' : 'router_pb2'
  # @@protoc_insertion_point(class_scope:Router.RouterUserAuth)
  })
_sym_db.RegisterMessage(RouterUserAuth)

RouterDeviceAuth = _reflection.GeneratedProtocolMessageType('RouterDeviceAuth', (_message.Message,), {
  'DESCRIPTOR' : _ROUTERDEVICEAUTH,
  '__module__' : 'router_pb2'
  # @@protoc_insertion_point(class_scope:Router.RouterDeviceAuth)
  })
_sym_db.RegisterMessage(RouterDeviceAuth)

RouterRecordRotation = _reflection.GeneratedProtocolMessageType('RouterRecordRotation', (_message.Message,), {
  'DESCRIPTOR' : _ROUTERRECORDROTATION,
  '__module__' : 'router_pb2'
  # @@protoc_insertion_point(class_scope:Router.RouterRecordRotation)
  })
_sym_db.RegisterMessage(RouterRecordRotation)

RouterRecordRotationsRequest = _reflection.GeneratedProtocolMessageType('RouterRecordRotationsRequest', (_message.Message,), {
  'DESCRIPTOR' : _ROUTERRECORDROTATIONSREQUEST,
  '__module__' : 'router_pb2'
  # @@protoc_insertion_point(class_scope:Router.RouterRecordRotationsRequest)
  })
_sym_db.RegisterMessage(RouterRecordRotationsRequest)

RouterRecordRotationsResponse = _reflection.GeneratedProtocolMessageType('RouterRecordRotationsResponse', (_message.Message,), {
  'DESCRIPTOR' : _ROUTERRECORDROTATIONSRESPONSE,
  '__module__' : 'router_pb2'
  # @@protoc_insertion_point(class_scope:Router.RouterRecordRotationsResponse)
  })
_sym_db.RegisterMessage(RouterRecordRotationsResponse)

RouterRotationInfo = _reflection.GeneratedProtocolMessageType('RouterRotationInfo', (_message.Message,), {
  'DESCRIPTOR' : _ROUTERROTATIONINFO,
  '__module__' : 'router_pb2'
  # @@protoc_insertion_point(class_scope:Router.RouterRotationInfo)
  })
_sym_db.RegisterMessage(RouterRotationInfo)

RouterRecordRotationRequest = _reflection.GeneratedProtocolMessageType('RouterRecordRotationRequest', (_message.Message,), {
  'DESCRIPTOR' : _ROUTERRECORDROTATIONREQUEST,
  '__module__' : 'router_pb2'
  # @@protoc_insertion_point(class_scope:Router.RouterRecordRotationRequest)
  })
_sym_db.RegisterMessage(RouterRecordRotationRequest)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.keepersecurity.protoB\006Router'
  _ROUTERRESPONSECODE._serialized_start=1350
  _ROUTERRESPONSECODE._serialized_end=1532
  _ROUTERROTATIONSTATUS._serialized_start=1534
  _ROUTERROTATIONSTATUS._serialized_end=1641
  _ROUTERRESPONSE._serialized_start=35
  _ROUTERRESPONSE._serialized_end=149
  _ROUTERCONTROLLERMESSAGE._serialized_start=152
  _ROUTERCONTROLLERMESSAGE._serialized_end=313
  _ROUTERUSERAUTH._serialized_start=316
  _ROUTERUSERAUTH._serialized_end=462
  _ROUTERDEVICEAUTH._serialized_start=465
  _ROUTERDEVICEAUTH._serialized_end=650
  _ROUTERRECORDROTATION._serialized_start=652
  _ROUTERRECORDROTATION._serialized_end=762
  _ROUTERRECORDROTATIONSREQUEST._serialized_start=764
  _ROUTERRECORDROTATIONSREQUEST._serialized_end=833
  _ROUTERRECORDROTATIONSRESPONSE._serialized_start=835
  _ROUTERRECORDROTATIONSRESPONSE._serialized_end=932
  _ROUTERROTATIONINFO._serialized_start=935
  _ROUTERROTATIONINFO._serialized_end=1167
  _ROUTERRECORDROTATIONREQUEST._serialized_start=1170
  _ROUTERRECORDROTATIONREQUEST._serialized_end=1347
# @@protoc_insertion_point(module_scope)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: commander@keepersecurity.com
#
# Example code to retrieve the password for a record
# stored in the vault.  This example also pulls configuration
# from config.json or writes the config file if it does not exist.
#
# Usage:
#    python shared_folder.py

import os
import json
import base64
import getpass

from keepercommander.params import KeeperParams
from keepercommander import api
from keepercommander.commands.register import ShareFolderCommand


def read_config_file(params):
    params.config_filename = os.path.join(os.path.dirname(__file__), 'config.json')
    if os.path.isfile(params.config_filename):
        with open(params.config_filename, 'r') as f:
            params.config = json.load(f)
            if 'user' in params.config:
                params.user = params.config['user']

            if 'password' in params.config:
                params.password = params.config['password']

            if 'mfa_token' in params.config:
                params.mfa_token = params.config['mfa_token']

            if 'server' in params.config:
                params.server = params.config['server']

            if 'device_id' in params.config:
                device_id = base64.urlsafe_b64decode(params.config['device_id'] + '==')
                params.rest_context.device_id = device_id


my_params = KeeperParams()
read_config_file(my_params)

while not my_params.user:
    my_params.user = getpass.getpass(prompt='User(Email): ', stream=None)

while not my_params.password:
    my_params.password = getpass.getpass(prompt='Master Password: ', stream=None)

api.sync_down(my_params)

folders = ["Shared Folder 1", "Shared Folder 2"]
user1 = "user@company.com"
team1 = "Team1"

command = ShareFolderCommand()
for folder in folders:
    command.execute(my_params, action="grant", user=[user1, team1], folder=folder)

api.sync_down(my_params)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import abc
import importlib
import io
import json
import logging
import os.path
from contextlib import contextmanager
from typing import List, Optional, Union, Dict, Iterable

from ..error import CommandError
from ..params import KeeperParams
from ..recordv3 import RecordV3

PathDelimiter = '\\'
TWO_FACTOR_CODE = 'TFC:Keeper'
FIELD_TYPE_ONE_TIME_CODE = 'oneTimeCode'


def replace_email_domain(email, old_domain, new_domain):
    # type: (str, Optional[str], Optional[str]) -> str
    if old_domain and new_domain and email.endswith(f'@{old_domain}'):
        email_user, domain = email.split('@')
        return f'{email_user}@{new_domain}'
    else:
        return email


def importer_for_format(input_format):
    full_name = 'keepercommander.importer.' + input_format
    module = importlib.import_module(full_name)
    if hasattr(module, 'Importer'):
        return module.Importer
    raise Exception('Cannot resolve importer for format {}'.format(input_format))


def exporter_for_format(output_format):
    full_name = 'keepercommander.importer.' + output_format
    module = importlib.import_module(full_name)
    if hasattr(module, 'Exporter'):
        return module.Exporter
    raise Exception('Cannot resolve exporter for format {}'.format(output_format))


def strip_path_delimiter(name, delimiter=PathDelimiter):
    folder = name.strip()
    if folder == delimiter:
        return ''
    if len(folder) > 1:
        if folder[:1] == delimiter and folder[:2] != delimiter*2:
            folder = folder[1:].strip()
    if len(folder) > 1:
        if folder[-1:] == delimiter and folder[-2:] != delimiter*2:
            folder = folder[:-1].strip()
    return folder


def path_components(path, delimiter=PathDelimiter):    # type: (str, str) -> Iterable[str]
    p = path.strip()
    pos = 0
    while pos < len(p):
        idx = p.find(delimiter, pos)
        if idx >= 0:
            if idx+1 < len(p):
                if p[idx+1] == delimiter:
                    pos = idx + 2
                    continue
            comp = p[:idx].strip()
            p = p[idx+1:].strip()
            pos = 0
            if len(comp) > 0:
                yield comp.replace(2*delimiter, delimiter)
        else:
            p = strip_path_delimiter(p, delimiter=delimiter)
            if len(p) > 0:
                yield p.replace(2*delimiter, delimiter)
                p = ''


def check_if_bool(value):
    return value is None or type(value) == bool


class Permission:
    def __init__(self):
        self.uid = None
        self.name = None
        self.manage_users = None
        self.manage_records = None


class SharedFolder:
    def __init__(self):
        self.uid = None
        self.path = None
        self.manage_users = None
        self.manage_records = None
        self.can_edit = None
        self.can_share = None
        self.permissions = None  # type: Optional[List[Permission]]

    def validate(self):
        if not self.path:
            raise CommandError('import', 'shared folder. path cannot be empty')
        for attr in ['manage_users', 'manage_records', 'can_edit', 'can_share']:
            if hasattr(self, attr):
                if not check_if_bool(getattr(self, attr)):
                    raise CommandError('import', 'shared folder. property \'{0}\' should be a boolean'.format(attr))
        if self.permissions:
            for p in self.permissions:
                if not p.name and not p.uid:
                    raise CommandError('import', 'shared folder permission. property \'name\' cannot be empty'.format(attr))
                for attr in ['manage_users', 'manage_records']:
                    if hasattr(p, attr):
                        if not check_if_bool(getattr(p, attr)):
                            raise CommandError('import', 'shared folder permission. property \'{0}\' should be a boolean'.format(attr))


class Team:
    def __init__(self):
        self.uid = None
        self.name = None
        self.members = None   # type: Optional[List[str]]

    def validate(self):
        if not self.name:
            raise CommandError('', 'Team name cannot be empty')


class Attachment(abc.ABC):
    def __init__(self):
        self.name = None
        self.size = None
        self.mime = None

    @abc.abstractmethod
    def open(self):  # type: () -> io.BufferedIOBase
        pass

    def prepare(self):   # type: () -> None
        pass


class Folder:
    def __init__(self):
        self.uid = None        # type: Optional[str]
        self.domain = None     # type: Optional[str]
        self.path = None       # type: Optional[str]
        self.can_edit = None   # type: Optional[str]
        self.can_share = None  # type: Optional[str]

    def get_folder_path(self):
        path = self.domain or ''
        if self.path:
            if path:
                path = path + PathDelimiter
            path = path + self.path
        return path


class RecordField:
    def __init__(self, type=None, label=None, value=None):
        self.type = type    # type: Optional[str]
        self.label = label   # type: Optional[str]
        self.value = value   # type: any

    def name_key(self):
        if self.type and self.label:
            return f'${self.type.lower()}.{self.label.lower()}'
        elif self.type:
            return f'${self.type.lower()}'
        else:
            return (self.label or '').lower()

    @staticmethod
    def hash_value(value):  # type: (any) -> str
        if not value:
            return ''
        if isinstance(value, str):
            value = value.strip()
        elif isinstance(value, list):
            value = [RecordField.hash_value(x) for x in value]
            value = '|'.join((x for x in value if x))
        elif isinstance(value, dict):
            keys = [x for x in value]
            if 'privateKey' in keys:
                if value['privateKey']:
                    try:
                        keys.remove('publicKey')
                    except:
                        pass
            keys.sort()
            kvp = [(x, RecordField.hash_value(value[x])) for x in keys]
            kvp = [x for x in kvp if x[1]]
            value = ';'.join((f'{x[0]}:{x[1]}' for x in kvp))
        else:
            value = str(value)
        return value

    def hash_key(self):  # type: () -> Optional[str]
        value = RecordField.hash_value(self.value)
        if value:
            name = self.name_key()
            return f'{name}:{value}'


class RecordReferences:
    def __init__(self, type='', label=None):
        self.type = type
        self.label = label
        self.uids = []   # type: List[any]


class Record:
    def __init__(self):
        self.uid = None
        self.type = None         # type: Optional[str]
        self.title = None
        self.login = None
        self.password = None
        self.login_url = None
        self.notes = None
        self.fields = []         # type: List[RecordField]
        self.folders = None      # type: Optional[List[Folder]]
        self.attachments = None  # type: Optional[List[Attachment]]
        self.references = None   # type: Optional[List[RecordReferences]]
        self.schema = None       # type: Optional[List[RecordSchemaField]]

    def validate(self):
        if not self.title:
            raise CommandError('import', 'Record: title cannot be empty')
        if self.folders:
            for f in self.folders:
                for attr in ['can_edit', 'can_share']:
                    if hasattr(f, attr):
                        if not check_if_bool(getattr(f, attr)):
                            raise CommandError('import', f'Record \'{self.title}\': folder property \'{attr}\' should be boolean')


class RecordSchemaField:
    def __init__(self):
        self.ref = ''
        self.label = ''


class File:
    def __init__(self):
        self.file_id = None
        self.title = None
        self.name = None
        self.size = None
        self.key = None
        self.mime = None

    def open(self):
        raise NotImplementedError()

    def validate(self):
        pass


class BaseImporter(abc.ABC):
    def execute(self, name, **kwargs):
        # type: (BaseImporter, str, dict) -> Iterable[Union[Record, SharedFolder, File]]
        yield from self.do_import(name, **kwargs)

    @abc.abstractmethod
    def do_import(self, filename, **kwargs):
        # type: (BaseImporter, str, dict) -> Iterable[Union[Record, SharedFolder, File]]
        pass

    def extension(self):
        return ''

    @staticmethod
    def import_host_field(value):    # type: (str) -> Optional[dict]
        if isinstance(value, str):
            host, _, port = value.partition(':')
            return {
                'hostName': host.strip(),
                'port': port.strip()
            }

    @staticmethod
    def import_phone_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            region = ''
            number = ''
            ext = ''
            phone_type, _, rest = value.partition(':')
            if not rest:
                rest = phone_type
                phone_type = ''
            comps = rest.strip().split(' ')
            for comp in comps:
                comp = comp.strip()
                if comp.isalpha():
                    if len(comp) == 2:
                        if not region:
                            region = comp
                    elif not phone_type:
                        phone_type = comp
                elif len(comp) >= 6:
                    if not number:
                        number = comp
                elif not ext:
                    ext = comp
            result = {
                'type': '',
                'region': '',
                'number': number.strip(),
                'ext': ext.strip()
            }
            phone_type = phone_type.strip()
            region = region.strip()
            if phone_type:
                result['type'] = phone_type
            if region:
                result['region'] = region
            return result

    @staticmethod
    def import_name_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            first = ''
            middle = ''
            last = ''
            comma_pos = value.find(',')
            if comma_pos >= 0:
                last = value[:comma_pos]
                rest = value[comma_pos+1:]
            else:
                space_pos = value.rfind(' ')
                if space_pos >= 0:
                    last = value[space_pos+1:]
                    rest = value[:space_pos]
                else:
                    last = value
                    rest = ''
            rest = rest.strip()
            if rest:
                space_pos = rest.rfind(' ')
                if space_pos >= 0:
                    middle = rest[space_pos+1:]
                    first = rest[:space_pos]
                else:
                    middle = ''
                    first = rest

            return {
                'first': first.strip(),
                'middle': middle.strip(),
                'last': last.strip(),
            }

    @staticmethod
    def import_address_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            comps = value.split(',')
            street1 = comps[0].strip() if len(comps) > 0 else ''
            city = comps[1].strip() if len(comps) > 1 else ''
            state, _, zip_code = comps[2].strip().partition(' ') if len(comps) > 2 else ('', '', '')
            country = comps[3].strip() if len(comps) > 3 else ''

            return {
                'street1': street1,
                'street2': '',
                'city': city,
                'state': state,
                'zip': zip_code,
                'country': country
            }

    @staticmethod
    def import_q_and_a_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            q, sign, a = value.partition('?')
            return {
                'question': q.strip() + '?',
                'answer': a.strip(),
            }

    @staticmethod
    def import_card_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            comps = value.split(' ')
            number = ''
            expiration = ''
            cvv = ''
            for comp in comps:
                comp = comp.strip()
                if comp:
                    if len(comp) > 10:
                        number = comp
                    elif comp.find('/') >= 0:
                        expiration = comp
                    elif len(comp) <= 6:
                        cvv = comp
            if number or expiration or cvv:
                return {
                    'cardNumber': number,
                    'cardExpirationDate': expiration,
                    'cardSecurityCode':  cvv,
                }

    @staticmethod
    def import_account_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            account_type = ''
            routing = ''
            account_number = ''
            comps = value.split()
            for comp in comps:
                comp = comp.strip()
                if comp.isnumeric():
                    if not routing:
                        routing = comp
                    elif not account_number:
                        account_number = comp
                else:
                    if not account_type:
                        account_type = comp
            if routing or account_number:
                return {
                    'accountType': account_type,
                    'routingNumber': routing,
                    'accountNumber': account_number
                }

    @staticmethod
    def import_ssh_key_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            return {
                'privateKey': value,
                'publicKey': ''
            }

    @staticmethod
    def import_field(field_type, field_value):  # type: (str, str) -> any
        if not field_value:
            return None
        if not field_type:
            return field_value
        if field_type in {'text', 'multiline', 'secret', 'note'}:
            return field_value
        if field_type == 'date':
            try:
                return int(field_value)
            except:
                return None
        if field_type == 'keyPair':
            return BaseImporter.import_ssh_key_field(field_value)

        str_values = field_value.split('\n')
        values = []
        for str_value in str_values:
            if field_type == 'host':
                v = BaseImporter.import_host_field(str_value)
            elif field_type == 'phone':
                v = BaseImporter.import_phone_field(str_value)
            elif field_type == 'name':
                v = BaseImporter.import_name_field(str_value)
            elif field_type == 'address':
                v = BaseImporter.import_address_field(str_value)
            elif field_type == 'securityQuestion':
                v = BaseImporter.import_q_and_a_field(str_value)
            elif field_type == 'paymentCard':
                v = BaseImporter.import_card_field(str_value)
            elif field_type == 'bankAccount':
                v = BaseImporter.import_account_field(str_value)
            else:
                v = str_value
                if field_type in RecordV3.field_values:
                    fv = RecordV3.field_values[field_type]
                    if isinstance(fv.get('value'), dict):
                        try:
                            v = json.loads(str_value)
                        except:
                            pass
                    elif isinstance(fv.get('value'), int):
                        try:
                            v = int(str_value)
                        except:
                            pass
            if v:
                values.append(v)
        if values:
            if len(values) == 1:
                return values[0]
            return values

    @staticmethod
    def adjust_field_label(record, field_type, field_label, fields):
        # type: (Record, str, str, List[Dict]) -> str
        if not isinstance(fields, list):
            return field_label
        if field_type == 'text':
            return field_label
        field = next((x for x in fields if x['$ref'] == field_type), None)
        if not field:
            return field_label
        fl = field.get('label', '')
        if fl == field_label:
            return field_label
        for f in record.fields:
            if f.type == field_type and f.label == fl:
                return field_label
        return fl


class BaseFileImporter(BaseImporter, abc.ABC):
    def __init__(self):
        super(BaseFileImporter, self).__init__()

    def execute(self, name, **kwargs):
        # type: (str, ...) -> Iterable[Union[Record, SharedFolder, File]]

        path = os.path.expanduser(name)
        if not os.path.isfile(path):
            ext = self.extension()
            if ext:
                path = path + '.' + ext

        if not os.path.isfile(path):
            raise CommandError('import', f'File \'{name}\' does not exist')

        yield from self.do_import(path, **kwargs)


class BaseExporter(abc.ABC):
    def __init__(self):
        self.max_size = 10 * 1024 * 1024

    def execute(self, filename, items, file_password=None):
        # type: (str, List[Union[Record, SharedFolder, File, Team]], Optional[str]) -> None
        if filename:
            filename = os.path.expanduser(filename)
            if filename.find('.') < 0:
                ext = self.extension()
                if ext:
                    filename = filename + '.' + ext
        elif not self.supports_stdout():
            raise CommandError('export', 'File name parameter is required.')

        self.do_export(filename, items, file_password)
        if filename and os.path.isfile(filename):
            logging.info('Vault has been exported to: %s', os.path.abspath(filename))

    @abc.abstractmethod
    def do_export(self, filename, records, file_password=None):
        # type: (str, List[Record, File], Optional[str]) -> None
        pass

    def has_shared_folders(self):
        return False

    def has_attachments(self):
        return False

    def extension(self):
        return ''

    def supports_stdout(self):
        return False

    def supports_v3_record(self):
        return True

    @staticmethod
    def export_host_field(value):    # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            host = value.get('hostName') or ''
            port = value.get('port') or ''
            if host or port:
                if port:
                    host += ':' + port
            return host

    @staticmethod
    def export_phone_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            phone = value.get('type') or ''
            if phone:
                phone += ':'
            region = value.get('region') or ''
            if region:
                if len(region) == 2 and region.isalpha():
                    pass
                elif region.isnumeric():
                    region = '+' + region
                else:
                    region = ''
                if region:
                    phone += '  ' + region
            number = (value.get('number') or '').replace(' ', '-')
            if number:
                phone += ' ' + number
            ext = (value.get('ext') or '').replace(' ', '-')
            if ext:
                phone += ' ' + ext
            return phone

    @staticmethod
    def export_name_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            first_name = value.get('first') or ''
            middle_name = value.get('middle') or ''
            name = value.get('last') or ''
            if first_name or middle_name or name:
                name = f'{name}, {first_name}'
                if middle_name:
                    name += ' ' + middle_name
            return name

    @staticmethod
    def export_address_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            address = value.get('street1', '').replace(',', '.')
            street2 = value.get('street2', '').replace(',', '.')
            if street2:
                address += ' ' + street2
            city = value.get('city', '').replace(',', '.')
            if city:
                address += ', ' + city
                state = value.get('state', '').replace(',', '.')
                zip_code = value.get('zip', '').replace(',', '.')
                if state or zip_code:
                    address += ', ' + state + ' ' + zip_code
                    country = value.get('country', '').replace(',', '.')
                    if country:
                        address += ', ' + country
            return address

    @staticmethod
    def export_q_and_a_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            q = value.get('question', '').replace('?', '')
            a = value.get('answer', '')
            if q or a:
                return f'{q}? {a}'
            return q

    @staticmethod
    def export_card_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            comps = []
            number = value.get('cardNumber')
            if number:
                comps.append(number)
            expiration = value.get('cardExpirationDate')
            if expiration:
                comps.append(expiration)
            cvv = value.get('cardSecurityCode')
            if cvv:
                comps.append(cvv)
            if comps:
                return ' '.join(comps)

    @staticmethod
    def export_account_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            account_type = value.get('accountType', '').replace(' ', '')
            routing = value.get('routingNumber', '').replace(' ', '')
            account_number = value.get('accountNumber', '').replace(' ', '')
            if routing or account_number:
                comps = []
                if account_type:
                    comps.append(account_type)
                if routing:
                    comps.append(routing)
                if account_number:
                    comps.append(account_number)
                if comps:
                    return ' '.join(comps)

    @staticmethod
    def export_ssh_key_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            return value.get('privateKey', '')

    @staticmethod
    def export_field(field_type, field_value):  # type: (str, any) -> str
        if not field_value:
            return ''

        if isinstance(field_value, str):
            return field_value
        if isinstance(field_value, list):
            values = []
            for value in field_value:
                v = BaseExporter.export_field(field_type, value)
                if v:
                    values.append(v)
            return '\n'.join((x.replace('\n', ' ') for x in values))
        if isinstance(field_value, dict):
            if field_type == 'host':
                return BaseExporter.export_host_field(field_value)
            if field_type == 'phone':
                return BaseExporter.export_phone_field(field_value)
            if field_type == 'name':
                return BaseExporter.export_name_field(field_value)
            if field_type == 'address':
                return BaseExporter.export_address_field(field_value)
            if field_type == 'securityQuestion':
                return BaseExporter.export_q_and_a_field(field_value)
            if field_type == 'paymentCard':
                return BaseExporter.export_card_field(field_value)
            if field_type == 'bankAccount':
                return BaseExporter.export_account_field(field_value)
            if field_type == 'keyPair':
                return BaseExporter.export_ssh_key_field(field_value)
            return json.dumps(field_value)

        return str(field_value)


class BytesAttachment(Attachment):
    def __init__(self, name, buffer):  # type: (str, bytes) -> None
        Attachment.__init__(self)
        self.name = name
        self.data = buffer
        self.size = len(buffer)

    @contextmanager
    def open(self):
        yield io.BytesIO(self.data)


class BaseDownloadMembership(abc.ABC):
    @abc.abstractmethod
    def download_membership(self, params):   # type: (KeeperParams) -> Iterable[Union[SharedFolder, Team]]
        pass
